from typing import List, Optional

import litellm

from rustic_ai.core.guild.agent_ext.depends.dependency_resolver import (
    DependencyResolver,
)
from rustic_ai.core.guild.agent_ext.depends.llm import LLM
from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionTool,
)
from rustic_ai.litellm.conf import LiteLLMConf
from rustic_ai.litellm.utils import ResponseUtils, transform_response_format


class LiteLLM(LLM):
    def __init__(self, props: LiteLLMConf):
        litellm.drop_params = True
        litellm.set_verbose = False
        # set callbacks
        litellm.input_callback = []
        litellm.success_callback = []
        litellm.failure_callback = []
        self.pre_messages = props.messages
        self._model = props.model
        self.client_props = props.model_dump(
            mode="json",
            exclude_unset=True,
            exclude_none=True,
            exclude={
                "message_memory",
                "toolset",
                "filter_attachments",
                "extract_tool_calls",
                "skip_chat_response_on_tool_call",
                "retries_on_tool_parse_error",
            },
        )

        self.tool_manager = props.get_tools_manager()

    def _prep_prompt(self, prompt: ChatCompletionRequest) -> dict:
        messages = self.pre_messages if self.pre_messages else []

        all_messages = messages + prompt.messages

        messages_dict = [m.model_dump(exclude_none=True) for m in all_messages]

        tools: List[ChatCompletionTool] = self.tool_manager.tools if self.tool_manager else []
        if prompt.tools:
            tools.extend(prompt.tools)

        prompt.response_format = transform_response_format(prompt.response_format)

        prompt_dict = prompt.model_dump(exclude_none=True, exclude_unset=True)

        full_prompt = {
            **self.client_props,
            **prompt_dict,
            "messages": messages_dict,
        }

        if tools:
            # Convert ChatCompletionTool objects to dictionaries for litellm
            full_prompt["tools"] = [tool.model_dump(exclude_none=True) for tool in tools]

        return full_prompt

    def completion(self, prompt: ChatCompletionRequest, model: Optional[str] = None):
        full_prompt = self._prep_prompt(prompt)

        if model:
            full_prompt["model"] = model

        completion = litellm.completion(**full_prompt)
        response: ChatCompletionResponse = ResponseUtils.from_litellm(completion)
        response.input_messages = full_prompt["messages"]  # Store input messages in the response
        return response

    async def async_completion(self, prompt: ChatCompletionRequest, model: Optional[str] = None):
        full_prompt = self._prep_prompt(prompt)

        if model:
            full_prompt["model"] = model

        completion = await litellm.acompletion(**full_prompt)
        response: ChatCompletionResponse = ResponseUtils.from_litellm(completion)
        response.input_messages = full_prompt["messages"]  # Store input messages in the response
        return response

    @property
    def model(self) -> str:
        return self._model

    def get_config(self) -> dict:
        return self.client_props


class LiteLLMResolver(DependencyResolver[LLM]):
    memoize_resolution: bool = False

    def __init__(self, model: str, conf: dict = {}):
        super().__init__()
        conf["model"] = model
        self.props = LiteLLMConf.model_validate(conf)
        self.LiteLLM = LiteLLM(self.props)

    def resolve(self, org_id: str, guild_id: str, agent_id: str) -> LLM:
        return self.LiteLLM  # We can always return the same instance of LiteLLM



from collections import deque
from datetime import datetime
import logging
from typing import Deque, List, Optional, Union

from dotenv import load_dotenv
import litellm
from litellm import validate_environment
import openai

from rustic_ai.core.guild.agent import (
    Agent,
    ProcessContext,
    processor,
)
from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    ArrayOfContentParts,
    AssistantMessage,
    ChatCompletionError,
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatCompletionResponseEmpty,
    ChatCompletionTool,
    FileContentPart,
    FunctionMessage,
    ImageContentPart,
    ResponseCodes,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from rustic_ai.core.guild.agent_ext.depends.llm.tools_manager import ToolsManager
from rustic_ai.litellm.utils import ResponseUtils, transform_response_format

from .conf import LiteLLMConf

load_dotenv()


class LiteLLMAgent(Agent[LiteLLMConf]):
    """
    LiteLLM agent for invoking various LLM models using LiteLLM.
    """

    litellm.drop_params = True

    def __init__(self):
        validation_result = validate_environment(self.config.model)
        if not validation_result.get("keys_in_environment"):
            missing_keys = validation_result.get("missing_keys", [])
            raise RuntimeError(f"Required environment variable(s) {'{}'.join(missing_keys)} not set")

        self.pre_messages = self.config.messages

        self.model = self.config.model
        self.client_props = self.config.model_dump(
            mode="json",
            exclude_unset=True,
            exclude_none=True,
            exclude={
                "message_memory",
                "toolset",
                "filter_attachments",
                "extract_tool_calls",
                "skip_chat_response_on_tool_call",
                "retries_on_tool_parse_error",
            },
        )
        self.message_memory_size = self.config.message_memory or 0
        self.message_queue: Deque[SystemMessage | UserMessage | AssistantMessage | ToolMessage | FunctionMessage] = (
            deque(maxlen=self.message_memory_size)
        )
        self.filter_attachments = self.config.filter_attachments
        self.tools_manager: Optional[ToolsManager] = self.config.get_tools_manager()

        self.extract_tool_calls = self.config.extract_tool_calls
        self.skip_chat_response_on_tool_call = self.config.skip_chat_response_on_tool_call
        self.retries_on_tool_parse_error = self.config.retries_on_tool_parse_error

    @processor(ChatCompletionRequest)
    def llm_completion(self, ctx: ProcessContext[ChatCompletionRequest]):

        prompt = ctx.payload
        prompt.response_format = transform_response_format(prompt.response_format)

        messages = self.pre_messages if self.pre_messages else []

        user_messages = prompt.messages
        send_messages = user_messages

        if user_messages and self.filter_attachments:
            # Filter out attachments from user messages
            filtered_messages = []
            for msg in user_messages:
                if isinstance(msg, UserMessage) and isinstance(msg.content, ArrayOfContentParts):
                    msg.content = [
                        content
                        for content in msg.content
                        if not isinstance(content, FileContentPart) and not isinstance(content, ImageContentPart)
                    ]
                    if msg.content:
                        filtered_messages.append(msg)
                else:
                    filtered_messages.append(msg)

            send_messages = filtered_messages

        if send_messages:
            all_messages = messages + list(self.message_queue) + send_messages

            messages_dict = [m.model_dump(exclude_none=True) for m in all_messages]

            tools: List[ChatCompletionTool] = self.tools_manager.tools if self.tools_manager else []
            if prompt.tools:
                tools.extend(prompt.tools)

            full_prompt = {
                **self.client_props,
                **prompt.model_dump(exclude_unset=True, exclude_none=True),
                "messages": messages_dict,
            }

            if tools:
                # Convert ChatCompletionTool objects to dictionaries for litellm
                full_prompt["tools"] = [tool.model_dump(exclude_none=True) for tool in tools]

            response = self._invoke_llm_and_process_response(ctx, full_prompt, all_messages)

            if response:

                # Update the message queue with the new messages
                self.message_queue.extend(
                    prompt.messages
                )  # Append the prompt messages (from user) to the message queue
                if response.choices and response.choices[0].message:
                    self.message_queue.append(
                        response.choices[0].message
                    )  # Append the response message (from LLM) to the message queue
        else:
            # No messages to send, return an error
            ctx.send(
                ChatCompletionResponseEmpty(
                    id=f"chatcmpl-{prompt.id}",
                    created=int(datetime.now().timestamp()),
                )
            )

    def _invoke_llm_and_process_response(
        self,
        ctx: ProcessContext[ChatCompletionRequest],
        full_prompt: dict,
        all_messages: List[Union[SystemMessage, UserMessage, AssistantMessage, ToolMessage, FunctionMessage]],
        call_count: int = 0,
    ) -> Optional[ChatCompletionResponse]:
        """
        Invoke the LLM and process the response.
        """
        try:
            completion = litellm.completion(**full_prompt)

            response: ChatCompletionResponse = ResponseUtils.from_litellm(completion)

            if response:
                if self.tools_manager and self.extract_tool_calls:
                    # Extract tool calls from the response
                    try:
                        tool_calls = self.tools_manager.extract_tool_calls(response)
                        if tool_calls:
                            # If tool calls are found, publish them
                            for tool_call in tool_calls:
                                ctx.send(tool_call)

                            # If skip_chat_response_on_tool_call is set, skip publishing the chat response
                            if self.skip_chat_response_on_tool_call:
                                return response  # We still return the response for tracking purposes
                    except Exception as e:
                        ctx.send_error(
                            ChatCompletionError(
                                status_code=ResponseCodes.INTERNAL_SERVER_ERROR,
                                message=f"Error extracting tool calls: {e}",
                                response=None,
                                model=self.model,
                                request_messages=all_messages,
                            )
                        )
                        # Retry the LLM completion if tool parse error occurs
                        # This is a workaround for the LLM returning invalid tool calls
                        if call_count <= self.retries_on_tool_parse_error:
                            call_count += 1
                            self._invoke_llm_and_process_response(
                                ctx=ctx,
                                full_prompt=full_prompt,
                                all_messages=all_messages,
                                call_count=call_count,
                            )
                        else:
                            ctx.send_error(
                                ChatCompletionError(
                                    status_code=ResponseCodes.INTERNAL_SERVER_ERROR,
                                    message="Max retries reached for tool parse error",
                                    response=None,
                                    model=self.model,
                                    request_messages=all_messages,
                                )
                            )
                            return None
                ctx.send(response)
            return response
        except openai.APIStatusError as e:  # pragma: no cover
            logging.error(f"Error in LLM completion: {e}")
            # Publish the error message
            self.process_api_status_error(
                ctx=ctx,
                status_code=ResponseCodes(e.status_code),
                error=e,
                messages=all_messages,
            )
            return None

    def process_api_status_error(
        self,
        ctx: ProcessContext[ChatCompletionRequest],
        status_code: ResponseCodes,
        error: openai.APIStatusError,
        messages: List[Union[SystemMessage, UserMessage, AssistantMessage, ToolMessage, FunctionMessage]],
    ):  # pragma: no cover
        """
        Process API status error and return a ChatCompletionError object.
        """
        ctx.send_error(
            ChatCompletionError(
                status_code=status_code,
                message=error.message,
                response=error.response.text if error.response else None,
                model=self.model,
                request_messages=messages,
            )
        )  # pragma: no cover



import logging
import re

from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    ChatCompletionRequest,
    ChatCompletionResponse,
)
from rustic_ai.core.messaging.core.message import AgentTag, FunctionalTransformer
from rustic_ai.core.utils.basic_class_utils import get_qualified_class_name


class LLMXformLambdas:

    _tag_pattern = re.compile(r"@([a-zA-Z0-9_-]+)")

    @staticmethod
    def ccres2ccreq(payload: dict) -> dict:
        """Converts a ChatCompletionRequest to a ChatCompletionResponse."""

        try:
            ccresp = ChatCompletionResponse.model_validate(payload)
            return {
                "payload": ChatCompletionRequest(messages=[ccresp.choices[0].message]).model_dump(),
                "format": get_qualified_class_name(ChatCompletionRequest),
            }

        except Exception as e:
            logging.error(f"Failed to validate payload: {e}")
            return {}

    @staticmethod
    def extract_agent_tags(payload: dict) -> dict:
        """Extracts agent tags from a ChatCompletionRequest."""

        try:
            ccresp = ChatCompletionResponse.model_validate(payload)
            message = ccresp.choices[0].message.content
            agent_ats = re.findall(LLMXformLambdas._tag_pattern, message)

            agent_tags = [AgentTag(id=tag) for tag in agent_ats]
            return {
                "recipient_list": agent_tags,
            }
        except Exception as e:
            logging.error(f"Failed to validate payload: {e}")
            return {}

    @classmethod
    def register_helpers(cls):
        """Registers the helper functions as lambdas."""
        FunctionalTransformer.register_lambda(
            name="ccres2ccreq",
            func=cls.ccres2ccreq,
        )
        FunctionalTransformer.register_lambda(
            name="extract_agent_tags",
            func=cls.extract_agent_tags,
        )



from typing import Annotated, Any, List, Literal, Optional, Type, Union, cast

from pydantic import BaseModel, ConfigDict, Field

from rustic_ai.core.guild.agent_ext.depends.llm.models import (
    AssistantMessage,
    FunctionMessage,
    Models,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from rustic_ai.core.guild.agent_ext.depends.llm.tools_manager import (
    Toolset,
    ToolsManager,
    ToolSpec,
)
from rustic_ai.core.guild.dsl import BaseAgentProps
from rustic_ai.core.utils.basic_class_utils import get_class_from_name


class ToolsetClassModel(BaseModel):
    type: Literal["toolset_class"] = Field("toolset_class", frozen=True)
    class_path: str  # will be resolved later


class ToolspecListModel(BaseModel):
    type: Literal["toolspec_list"] = Field("toolspec_list", frozen=True)
    tools: List[ToolSpec]


ToolsetUnion = Annotated[Union[ToolsetClassModel, ToolspecListModel], Field(discriminator="type")]


class LiteLLMConf(BaseAgentProps):
    """
    Properties to apply to all LiteLLM calls made by the agent.
    """

    model_config = ConfigDict(extra="ignore")

    model: Annotated[Union[str, Models], Field(examples=["gpt-4-turbo"])]
    """
    ID of the model to use. See the [model endpoint compatibility](/docs/models/model-endpoint-compatibility) table
    for details on which models work with the Chat API.
    """
    stream: Optional[bool] = False
    """
    If set, partial message deltas will be sent, like in ChatGPT. Tokens will be sent as data-only
    [server-sent events](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#Event_stream_format)
    as they become available, with the stream terminated by a `data: [DONE]` message.
    [Example Python code](https://cookbook.openai.com/examples/how_to_stream_completions).

    """
    # set api_base, api_version, api_key
    base_url: Optional[str] = None
    api_version: Optional[str] = None

    """
    If provided, return a mock completion response for testing or debugging purposes (default is None).
    """

    custom_llm_provider: Optional[str] = None
    """
    Used for Non-OpenAI LLMs, Example usage for bedrock, set model="amazon.titan-tg1-large" and custom_llm_provider="bedrock"
    """

    timeout: Optional[float] = None
    """
    The maximum time in seconds to wait for the completion.
    If the completion is not returned within this time, the request will time out.
    """

    messages: Optional[
        List[
            Union[
                SystemMessage,
                UserMessage,
                AssistantMessage,
                ToolMessage,
                FunctionMessage,
            ]
        ]
    ] = None
    """
    List of messages to send to the model before the prompt.
    These messages will be prependended to the list of messages in the prompt.
    """

    toolset: Optional[ToolsetUnion] = Field(
        default=None, description="Either {'type':'toolset_class', ...} or {'type':'toolspec_list', ...}"
    )
    """
    Toolset is an predefined implementation of Toolset (fully qualified name) or a list of tool
    specifications. Toolset gives a ToolManager which is used to work with tools and parse tool
    calls. The list of ToolSpec is used to create a ToolManager.
    """

    message_memory: Optional[int] = None
    """
    Number of messages to remember in the conversation history.
    """

    filter_attachments: bool = False
    """
    If set to True, attachments will be filtered out from the messages before sending to the model.
    """

    extract_tool_calls: bool = False
    """
    If set to True, the agent will extract tool calls from the model's response and
    publish them as well as chat completion responses.
    """

    skip_chat_response_on_tool_call: bool = False
    """
    If set to True, the agent will skip publishing the chat completion response when the response includes a tool call.
    Only the tool call will be published. This only applies when `extract_tool_calls` is set to True.
    """

    retries_on_tool_parse_error: int = 0
    """
    LLM may return an invalid tool call or with wrong args. In those cases, parsing will fail.
    Number of retries to LLM completion on tool parse error.
    If set to 0, the agent will not retry on tool parse error.
    """

    vertex_location: Optional[str] = None
    """
    Place where vertex model is deployed (us-central1, asia-southeast1, etc.). If None, attempts to use the VERTEXAI_LOCATION environment variable.
    """

    vertex_project: Optional[str] = None
    """
    The Google Cloud project ID. If None, attempts to use the VERTEXAI_PROJECT environment variable.
    """

    enable_json_schema_validation: bool = False
    """
    If true, enables json schema validation for all requests.
    """

    def get_tools_manager(self) -> Optional[ToolsManager]:
        """
        Returns the tools manager for the agent.
        If the toolset is a list of ToolSpec, it will create a new ToolsManager instance.
        If the toolset is a Toolset class, it will return the ToolsManager instance from the Toolset class.
        """
        if self.toolset is None:
            return None

        if isinstance(self.toolset, ToolsetClassModel):
            cls = self.path_to_cls(self.toolset.class_path)  # resolve string → subclass
            return cls.toolsmanager()  # class method supplied by Toolset

        if isinstance(self.toolset, ToolspecListModel):
            return ToolsManager(self.toolset.tools)

        raise TypeError("Unsupported toolset variant")

    @staticmethod
    def path_to_cls(value: str) -> Type[Toolset]:
        """
        Converts a string representation of a class back to the class itself.
        """
        if isinstance(value, type) and issubclass(value, Toolset):
            return cast(Type[Toolset], value)

        if not isinstance(value, str):
            raise TypeError(
                "parameter_class must be a fully-qualified string or a "
                "subclass of Toolset; got "
                f"{type(value).__name__}"
            )

        module_path, _, qualname = value.rpartition(".")
        if not module_path:
            raise ValueError(f"Not a fully-qualified path: {value!r}")

        mod = get_class_from_name(module_path)
        obj: Any = mod

        for attr in qualname.split("."):
            obj = getattr(obj, attr)

        if not isinstance(obj, type):
            raise TypeError(f"{value!r} does not resolve to a class")

        if not issubclass(obj, Toolset):
            raise TypeError(f"{value!r} is not a Toolset")

        return cast(Type[Toolset], obj)



from typing import Optional, Union

from litellm.utils import ModelResponse as LitellmModelResponse
from pydantic import BaseModel

from rustic_ai.core.guild.agent_ext.depends.llm.models import ChatCompletionResponse
from rustic_ai.core.utils.basic_class_utils import get_class_from_name
from rustic_ai.core.utils.json_utils import JsonDict


class ResponseUtils:
    @staticmethod
    def from_litellm(litellm_response: LitellmModelResponse) -> ChatCompletionResponse:
        response_dict = litellm_response.model_dump()
        response = ChatCompletionResponse.model_validate(response_dict)

        return response


def transform_response_format(value: Optional[Union[dict, str]]):
    if value is None:
        return None
    result: JsonDict = {
        "type": "json_schema",
        "json_schema": {"schema": {}, "strict": True},
    }
    if isinstance(value, dict):
        result["json_schema"]["schema"] = value
        return result
    if isinstance(value, str):
        format_type = get_class_from_name(value)
        # If it's a Pydantic model class, convert to JSON schema
        if isinstance(format_type, type) and issubclass(format_type, BaseModel):
            result["json_schema"]["schema"] = format_type.model_json_schema()
            return result
    return None
